/**
 * @file	cf_thread.c
 * @author	myusgun <myusgun@gmail.com>
 */
#include "cf_thread.h"
#include "cf_local.h"
#include "cf_error.h"

#include <stdlib.h>

#if defined(_WIN32) || defined(_WIN64)
# include <windows.h>
# include <process.h>
# define THREAD_TYPE	HANDLE
# define MUTEX_TYPE		HANDLE
# define THREAD_RETURN	unsigned long	/**< 스레드 워커 함수 반환 형 */
# define THREAD_CALL	__stdcall
#else // #if defined(_WIN32) || defined(_WIN64)
# include <pthread.h>
# define THREAD_TYPE	pthread_t
# define MUTEX_TYPE		pthread_mutex_t
# define THREAD_RETURN	void *			/**< 스레드 워커 함수 반환 형 */
# define THREAD_CALL
#endif // #if defined(_WIN32) || defined(_WIN64)

#define ASSERT_THREAD(__h)	\
	if (__h == NULL)		\
		return CF_ERROR_THREAD_INVALID_ARGS

#define ASSERT_MUTEX(__h)	\
	if (__h == NULL)		\
		return CF_ERROR_MUTEX_INVALID_ARGS

typedef THREAD_RETURN (THREAD_CALL * THREAD_WORKER) (void *);

static int
CF_Thread_Local_Close (void * ctx)
{
#if defined(_WIN32) || defined(_WIN64)
	CloseHandle (ctx);
#else
	free (ctx);
#endif

	return CF_OK;
}

/**
 * 스레드를 생성
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param threadID	스레드 아이디 주소
 * @param callback	스레드 워커 함수 이름
 * @param arg		스레드 함수로 전달할 인자
 *
 * @remarks
 * pthread(POSIX Thread)에서 지원되는 스케줄링 정책은 SCHED_OTHER, SCHED_FIFO, SCHED_RR 등이 존재 <br />
 * 일반적으로 설정되는 스케줄링 정책의 기본값은 SCHED_OTHER이며, 솔라리스 환경에서 SCHED_OTHER는 TS(timesharing) 방식으로 명시되어 있음 <br />
 * 그러나 개발 단계에서 테스트된 동작은 SCHED_FIFO와 동일하였으며, 때문에 솔라리스 환경에서는 스케줄링 정책을 SCHED_RR로 명시하도록 함 <br />
 * <br />
 * 참고 url <br />
 *  - http://kldp.org/node/18841 , "SCHED_OTHER, SCHED_FIFO, SCHED_RR에 대해..." <br />
 *  - http://blog.naver.com/popjunior/80021646476 , "AIX, CPU 모니터링과 튜닝" <br />
 */
int
CF_Thread_Create (CF_Thread				* threadID,
				  CF_Thread_Function	callback,
				  void					* arg)
{
	int result = 0;

	THREAD_WORKER f = (THREAD_WORKER) callback;

#if defined(_WIN32) || defined(_WIN64)
	*threadID = (THREAD_TYPE) CreateThread (NULL, 0, f, arg, 0, NULL);
	if (*threadID == NULL)
		return CF_ERROR_THREAD_CREATE;
#else

	pthread_attr_t		* attr = NULL;
# if defined(_SOLARIS)
	pthread_attr_t		solarisAttr;
	struct sched_param	sched;

	attr = &solarisAttr;

	result = pthread_attr_init (attr);
	if (result)
		return CF_ERROR_THREAD_INIT_ATTR;

	result = pthread_attr_setinheritsched (attr, PTHREAD_EXPLICIT_SCHED);
	if (result)
		return CF_ERROR_THREAD_SET_INHERIT_SCHED;

	result = pthread_attr_setschedpolicy (attr, SCHED_RR);
	if (result)
		return CF_ERROR_THREAD_SET_SCHED_POLICY;

	sched.sched_priority = 1;
	result = pthread_attr_setschedparam (attr, &sched);
	if (result)
		return CF_ERROR_THREAD_SET_SCHED_PARAM;
# endif // # if defined(_SOLARIS)

	*threadID = (THREAD_TYPE *) calloc (sizeof (THREAD_TYPE), 1);
	if (*threadID == NULL)
		return CF_ERROR_THREAD_ALLOC;

	result = pthread_create ((THREAD_TYPE *) *threadID, attr, f, arg);
	if (result)
	{
		CF_Thread_Release (threadID);
		return CF_ERROR_THREAD_CREATE;
	}

#endif
	return result;
}

/**
 * 스레드 아이디를 해제
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param threadID 스레드 아이디 주소
 *
 * @remark 스레드 아이디를 해제하는 것이며 워커 스레드가 종료되지 않음
 */
int
CF_Thread_Release (CF_Thread * threadID)
{
	ASSERT_THREAD (*threadID);

	return CF_Thread_Local_Close (*threadID);
}

/**
 * 스레드가 종료될 때 까지 대기
 *
 * @return CF_OK 반환
 *
 * @param threadID 스레드 아이디 주소
 */
int
CF_Thread_Join (CF_Thread * threadID)
{
	ASSERT_THREAD (*threadID);

#if defined(_WIN32) || defined(_WIN64)
	WaitForSingleObject ((THREAD_TYPE) *threadID, INFINITE);
#else
	char status[16] = {0x00,};
	pthread_join (*((THREAD_TYPE *) *threadID), (void **)status);
#endif

	return CF_OK;
}

/**
 * 뮤텍스 생성
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 *
 * @see CF_Thread_Create
 */
int
CF_Mutex_Create (CF_Mutex * mutex)
{
	int result = 0;

#if defined(_WIN32) || defined(_WIN64)
	*mutex = (MUTEX_TYPE) CreateMutexA (NULL, FALSE, NULL);
	if (*mutex == NULL)
		return CF_ERROR_MUTEX_CREATE;
#else
	*mutex = (MUTEX_TYPE *) calloc (sizeof (MUTEX_TYPE), 1);
	if (*mutex == NULL)
		return CF_ERROR_MUTEX_CREATE;

	result = pthread_mutex_init (*mutex, NULL);
	if (result)
		return CF_ERROR_MUTEX_CREATE;
#endif

	return CF_OK;
}

/**
 * 뮤텍스 해제
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 */
int
CF_Mutex_Destory (CF_Mutex * mutex)
{
	ASSERT_MUTEX (*mutex);

#if defined(_WIN32) || defined(_WIN64)
#else
	pthread_mutex_destroy (*mutex);
#endif
	return CF_Thread_Local_Close (*mutex);
}

/**
 * 뮤텍스 잠금
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 */
int
CF_Mutex_Lock (CF_Mutex * mutex)
{
	ASSERT_MUTEX (*mutex);

#if defined(_WIN32) || defined(_WIN64)
	WaitForSingleObject ((MUTEX_TYPE) *mutex, INFINITE);
#else
	pthread_mutex_lock ((MUTEX_TYPE *) *mutex);
#endif

	return CF_OK;
}

/**
 * 뮤텍스 잠금 해제
 *
 * @return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * @param mutex 뮤텍스 아이디 주소
 */
int
CF_Mutex_Unlock (CF_Mutex * mutex)
{
	ASSERT_MUTEX (*mutex);

#if defined(_WIN32) || defined(_WIN64)
	ReleaseMutex (*mutex);
#else
	pthread_mutex_unlock ((MUTEX_TYPE *) *mutex);
#endif

	return CF_OK;
}
